using ImmichToSlideshow.Models;
using ImmichToSlideshow.Models.Immich;
using Npgsql;
using System.Collections.ObjectModel;
using System.Data;
using System.Text;
using System.Text.Json;

namespace ImmichToSlideshow.Services;

public class AssetService(ILogger<Program> logger, AppSettings appSettings)
{

#pragma warning disable CS9124
    private readonly ILogger<Program> _Logger = logger;
    private readonly AppSettings _AppSettings = appSettings;
#pragma warning restore CS9124

    private static int? ExecuteNonQuery(string connectionString, string commandText, ReadOnlyCollection<NpgsqlParameter> npgsqlParameters)
    {
        int? result;
        if (string.IsNullOrEmpty(connectionString))
            result = null;
        else
        {
            using NpgsqlConnection npgsqlConnection = new(connectionString);
            npgsqlConnection.Open();
            using NpgsqlCommand npgsqlCommand = new(commandText, npgsqlConnection);
            npgsqlCommand.Parameters.AddRange(npgsqlParameters.ToArray());
            result = npgsqlCommand.ExecuteNonQuery();
        }
        return result;
    }

    private static StringBuilder GetForJsonPath(string connectionString, string commandText, ReadOnlyCollection<NpgsqlParameter> npgsqlParameters)
    {
        StringBuilder result = new();
        using NpgsqlConnection npgsqlConnection = new(connectionString);
        npgsqlConnection.Open();
        using NpgsqlCommand npgsqlCommand = new(commandText, npgsqlConnection);
        npgsqlCommand.Parameters.AddRange(npgsqlParameters.ToArray());
        NpgsqlDataReader npgsqlDataReader = npgsqlCommand.ExecuteReader(CommandBehavior.SequentialAccess);
        while (npgsqlDataReader.Read())
            _ = result.Append(npgsqlDataReader.GetString(0));
        return result;
    }

    public string? GetColumns()
    {
        string result;
        NpgsqlParameter[] npgsqlParameters = [];
        string commandText = CommandText.GetColumns();
        StringBuilder stringBuilder = GetForJsonPath(_AppSettings.ConnectionString, commandText, npgsqlParameters.AsReadOnly());
        result = stringBuilder.ToString();
        if (result.Length == 1)
            File.WriteAllText(".vscode/jsonl/.jsonl", result);
        return result;
    }

    public string? GetOwnerIds()
    {
        string result;
        NpgsqlParameter[] npgsqlParameters = [];
        string commandText = CommandText.GetOwnerIdActiveImage();
        StringBuilder stringBuilder = GetForJsonPath(_AppSettings.ConnectionString, commandText, npgsqlParameters.AsReadOnly());
        result = stringBuilder.ToString();
        if (result.Length == 1)
            File.WriteAllText(".vscode/jsonl/.jsonl", result);
        return result;
    }

    public string? GetAssets(Guid ownerId)
    {
        string result;
        NpgsqlParameter[] npgsqlParameters = [new NpgsqlParameter(nameof(ownerId), ownerId)];
        string commandText = CommandText.GetAssetActiveImagePreviewNotDuplicate(_AppSettings.FilterTags);
        StringBuilder stringBuilder = GetForJsonPath(_AppSettings.ConnectionString, commandText, npgsqlParameters.AsReadOnly());
        result = stringBuilder.ToString();
        if (result.Length == 1)
            File.WriteAllText(".vscode/jsonl/assets.jsonl", result);
        return result;
    }

    public ReadOnlyCollection<string>? GetRandomPaths(Guid ownerId)
    {
        string[]? results;
        NpgsqlParameter[] npgsqlParameters = [new NpgsqlParameter(nameof(ownerId), ownerId)];
        string commandText = CommandText.GetAssetActiveImagePreviewNotDuplicate(_AppSettings.FilterTags);
        StringBuilder stringBuilder = GetForJsonPath(_AppSettings.ConnectionString, commandText, npgsqlParameters.AsReadOnly());
        string json = stringBuilder.ToString();
        Random random = new();
        string ownerIdValue = ownerId.ToString();
        Asset[]? assets = JsonSerializer.Deserialize(json, AssetCollectionSourceGenerationContext.Default.AssetArray);
        results = assets is null ? null : (from l in assets orderby random.NextSingle() select l.Path.Split(ownerIdValue)[1]).ToArray();
        return results?.AsReadOnly();
    }

    public Tag[]? GetArchivedTag(AppSettings appSettings, Guid ownerId)
    {
        Tag[] results;
        Guid userId = ownerId;
        string value = appSettings.ArchivedTag;
        NpgsqlParameter[] npgsqlParameters =
        [
            new NpgsqlParameter(nameof(value), value),
            new NpgsqlParameter(nameof(userId), userId),
        ];
        string commandText = CommandText.GetArchivedTag();
        StringBuilder stringBuilder = GetForJsonPath(appSettings.ConnectionString, commandText, npgsqlParameters.AsReadOnly());
        string json = stringBuilder.ToString();
        results = JsonSerializer.Deserialize(json, TagCollectionSourceGenerationContext.Default.TagArray);
        return results;
    }

    public string? GetArchivedTag(Guid ownerId)
    {
        string result;
        Tag[]? tags = GetArchivedTag(_AppSettings, ownerId);
        result = tags is null || tags.Length != 1 ? null : tags[0].Id;
        return result;
    }

    public string? SaveRandomPaths(Guid ownerId)
    {
        string? results;
        DateTime dateTime = DateTime.Now;
        string tomorrow = dateTime.AddDays(1).ToString("MM-dd");
        if (Directory.Exists(_AppSettings.RandomResultsDirectory))
            _ = Directory.CreateDirectory(_AppSettings.RandomResultsDirectory);
        FileInfo fileInfo = new(Path.Combine(_AppSettings.RandomResultsDirectory, $"{tomorrow}.json"));
        if (fileInfo.Exists && fileInfo.CreationTime > dateTime.AddDays(_AppSettings.AddDays))
            results = null;
        else
        {
            _Logger.LogDebug("Writing <{FullName}>", fileInfo.FullName);
            ReadOnlyCollection<string>? paths = GetRandomPaths(ownerId);
            if (paths is null)
                results = null;
            else
            {
                _Logger.LogInformation("{count} path(s)", paths.Count.ToString());
                results = JsonSerializer.Serialize(paths);
                File.WriteAllText(fileInfo.FullName, results);
            }
        }
        return results;
    }

    private record Record(string Source, string Destination);

    public ReadOnlyCollection<string> SyncImmich(Guid ownerId)
    {
        List<string> results = [];
        Record record;
        List<Record> records = [];
        if (Directory.Exists(_AppSettings.SyncDirectory))
            _ = Directory.CreateDirectory(_AppSettings.SyncDirectory);
        int syncLength = _AppSettings.SyncDirectory.Length;
        string[] syncFiles = Directory.GetFiles(_AppSettings.SyncDirectory, "*", SearchOption.AllDirectories);
        string[] syncCheck = syncFiles.Select(l => l[syncLength..]).ToArray();
        if (Directory.Exists(_AppSettings.ImmichUploadDirectory))
            _ = Directory.CreateDirectory(_AppSettings.ImmichUploadDirectory);
        int immichUploadLength = _AppSettings.ImmichUploadDirectory.Length;
        string[] immichUploadFiles = Directory.GetFiles(_AppSettings.ImmichUploadDirectory, "*", SearchOption.AllDirectories);
        string[] immichUploadCheck = immichUploadFiles.Select(l => l[immichUploadLength..]).ToArray();
        for (int i = 0; i < immichUploadFiles.Length; i++)
        {
            if (syncCheck.Contains(immichUploadCheck[i]))
                continue;
            results.Add(immichUploadCheck[i]);
            record = new(immichUploadFiles[i], string.Concat(_AppSettings.SyncDirectory, immichUploadCheck[i]));
            records.Add(record);
        }
        _Logger.LogInformation("{count} file(s)", results.Count.ToString());
        for (int i = 0; i < records.Count; i++)
        {
            record = records[i];
            _Logger.LogDebug("Copying <{source}>", record.Source);
            File.Copy(record.Source, record.Destination);
        }
        return results.AsReadOnly();
    }

    public ReadOnlyCollection<int> SetArchiveImmich(Guid ownerId)
    {
        ReadOnlyCollection<int>? results;
        string checkDirectory = Path.Combine(_AppSettings.CurrentResultsDirectory, "B)Metadata", _AppSettings.CurrentCommit, "[]");
        if (!Directory.Exists(checkDirectory))
            results = null;
        else
        {
            string checkFile = Path.Combine(checkDirectory, "!9.json");
            if (!File.Exists(checkFile))
                results = null;
            else
            {
                string json = File.ReadAllText(checkFile);
                Identifier[]? identifiers = JsonSerializer.Deserialize<Identifier[]>(json);
                if (identifiers is null || identifiers.Length == 0)
                    results = null;
                else
                {
                    Tag[]? tags = GetArchivedTag(_AppSettings, ownerId);
                    if (tags is null || tags.Length != 1)
                        results = null;
                    else
                        results = SetArchiveImmich(logger, appSettings, ownerId, identifiers.AsReadOnly(), tags.AsReadOnly());
                }
            }
        }
        return results?.AsReadOnly();
    }

    private static ReadOnlyCollection<int> SetArchiveImmich(ILogger<Program> logger, AppSettings appSettings, Guid ownerId, ReadOnlyCollection<Identifier> identifiers, ReadOnlyCollection<Tag> tags)
    {
        ReadOnlyCollection<int>? results;
        string tagsId = tags[0].Id;
        NpgsqlParameter[] npgsqlParameters =
        [
            new NpgsqlParameter(nameof(tagsId), tagsId),
            new NpgsqlParameter(nameof(ownerId), ownerId),
        ];
        string deviceAssetIds = Identifier.GetDeviceAssetIds(identifiers);
        string commandText = CommandText.SetAssetArchived(deviceAssetIds);
        logger.LogDebug(commandText.Replace($"@{nameof(tagsId)}", $"'{tagsId}'").Replace($"@{nameof(ownerId)}", $"'{ownerId}'".ToString()));
        int? result = ExecuteNonQuery(appSettings.ConnectionString, commandText, npgsqlParameters.AsReadOnly());
        results = result is null ? null : new([result.Value]);
        return results?.AsReadOnly();
    }

}